/* Donated the the public domain 2012, 2015 Sam Trenholme.  Written in 2009,
 * 2012, and 2015. 
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

// This is actually a RG one-bit variant.  Actually, since the mill is
// 11 bits wide, it's better seen as a Gnoblio variant

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

/* This determines the word size we use for this particular Radio Gatun
 * implementation; DWR_WORDSIZE needs to be a multiple of 8 */
#define DWR_WORD uint32_t
#define DWR_WORDSIZE 32

// Pack word
#define PACK_WORD uint32_t
#define STATE_SIZE 67108864

/* These are hard coded in the Radio Gatun specification */
#define DWR_MILLSIZE 11
#define DWR_BELTROWS 3
#define DWR_BELTCOL 5
#define DWR_BELTFEED 4
#define DWR_BLANKROUNDS 16

int mill1 = 1;
int mill2 = 2;
/* The "mill" part of Radio Gatun */
void dwr_mill(DWR_WORD *a) {
	DWR_WORD A[DWR_MILLSIZE];
	DWR_WORD x;
	int i = 0;
	int y = 0;
	int r = 0;
	int z = 0;
	int q = 0;
	for(i = 0; i < DWR_MILLSIZE ; i++) {
		y = (i * 7) % DWR_MILLSIZE;
		r = ((i * (i + 1)) / 2) % DWR_WORDSIZE;
		x = a[y] ^ (a[ ((y + mill1) % DWR_MILLSIZE) ] | 
		    (~a[ ((y + mill2) % DWR_MILLSIZE) ]));
		//A[i] = (x >> r) | (x << (DWR_WORDSIZE - r));
		A[i] = x; // 1-bit, no rotate
	}
	for(i = 0; i < DWR_MILLSIZE ; i++) {
		y = i;
		z = (i + 1) % DWR_MILLSIZE;
		q = (i + 4) % DWR_MILLSIZE;
		a[i] = A[y] ^ A[z] ^ A[q];
	}
	a[0] ^= 1;
}	

/* The "belt" part of Radio Gatun */
void dwr_belt(DWR_WORD *a, DWR_WORD *b) {
	DWR_WORD q[DWR_BELTROWS];
	int s = 0;
	int i = 0;
	int v = 0;
	for(s = 0; s < DWR_BELTROWS ; s++) {
		q[s] = b[((s * DWR_BELTCOL) + DWR_BELTCOL - 1)];
	}
	for(i = DWR_BELTCOL - 1; i > 0; i--) {
		for(s = 0; s < DWR_BELTROWS ; s++) {
			v = i - 1;
			if(v < 0) {
				v = DWR_BELTCOL - 1;
			}
			b[((s * DWR_BELTCOL) + i)] = 
				b[((s * DWR_BELTCOL) + v)];
		}
	}
	for(s = 0; s < DWR_BELTROWS; s++) {
		b[(s * DWR_BELTCOL)] = q[s];
	}
	for(i = 0; i < DWR_BELTFEED ; i++) {
		s = (i + 1) + ((i % DWR_BELTROWS) * DWR_BELTCOL);
		b[s] ^= a[(i + 1)];
	}
	dwr_mill(a);
	for(i = 0; i < DWR_BELTROWS; i++) {
		a[(i + DWR_BELTCOL)] ^= q[i];
	}
}

// Convert the tiny belt + mill state of the RG/Gnoblio variant in
// to a single 32-bit number
PACK_WORD pack(DWR_WORD *belt, DWR_WORD *mill) {
	PACK_WORD out = 0;
	int a = 0;
	for(a=0;a<(DWR_BELTROWS * DWR_BELTCOL);a++) {
		out <<=1;
		out |= (belt[a] & 1);
	}	
	for(a=0;a<DWR_MILLSIZE;a++) {
		out <<=1;
		out |= (mill[a] & 1);
	}
	return out;
}

// Convert a packed number in to a belt + mill state
void unpack(PACK_WORD in, DWR_WORD *belt, DWR_WORD *mill) {
	int a = 0;
	for(a=DWR_MILLSIZE-1;a>=0;a--) {
		mill[a] = (in & 1);
		in >>= 1;
	}
	for(a=(DWR_BELTROWS * DWR_BELTCOL)-1;a>=0;a--) {
		belt[a] = (in & 1);
		in >>= 1;
	}
}

// Since I only have 16 gigs of memory -- blame Intel for the small
// memory limitation -- we need to pack the bits to look at the
// 34-bit 7/13 RG variant
int is_seen(uint8_t *seen, PACK_WORD this) {
	int bit, out;
	bit = 1 << (this & 7);
	this >>= 3;
	out = (seen[this] & bit);
	if(out != 0) {
		return 1;
	}
	return 0;
}

void set_seen(uint8_t *seen, PACK_WORD this) {
	int bit;
	bit = 1 << (this & 7);
	this >>= 3;
	seen[this] |= bit;
}

uint8_t *init_seen() {
	uint8_t *seen;
	PACK_WORD a;
	// Seen is too big for the Windows stack.  Ugh
	seen = malloc((STATE_SIZE >> 3) + 1);
	if(seen == 0) {
		puts("Malloc error");
		exit(1);
	}
	puts("Init seen array...");
	for(a=0;a<(STATE_SIZE >> 3);a++) {
		seen[a]=0;
	}
	puts("done...");
	return seen;
}
	
	
// Look at all of the cycle lengths for given Gamma constants m1 and m2
int see_cycles(int m1, int m2) {
	DWR_WORD belt[DWR_BELTROWS * DWR_BELTCOL];
	DWR_WORD mill[DWR_MILLSIZE];
	uint8_t *seen; // All possible states
	PACK_WORD a,b;	
	PACK_WORD this;
	int c = 0;

	seen = init_seen();
	mill1 = m1;
	mill2 = m2; 

	for(a=0;a<(DWR_BELTROWS * DWR_BELTCOL);a++) {
		belt[a] = 0;
	}
	for(a=0;a<DWR_MILLSIZE;a++) {
		mill[a] = 0;
	}
	for(a=b=0;b<STATE_SIZE;) {
		this = pack(belt,mill);
		if(b % 1000000 == 0) {
			printf("%d\n",b);
		}
		if(is_seen(seen,this) != 0) {
			printf("Repeats after %d nums at %08x\n",a,this);
			c++;
			if(c>=100) {
				free(seen);
				return 2;
			}
			a = 0;
			for(this=0;;this++) {
				if(is_seen(seen,this) == 0) {
					unpack(this,belt,mill);
					break;
				}
			}
		} else {
			a++;
			b++;
			set_seen(seen,this);
			dwr_belt(mill,belt);
		}
	}	
	free(seen);
	return 0;
}
	
int main(int argc, char **argv) {
	int a, b, c;
	for(a=0;a<=DWR_MILLSIZE;a++) {
		for(b=0;b<=DWR_MILLSIZE;b++) {
			c = see_cycles(a,b);
			if(c == 0) {
				printf("Mill constants: %d %d\n",a,b);
			}
			puts("");
		}
	}
	return 0;
}
